#!/bin/bash
#
# Basic script that is needed so that the bare minimum works when the
# package is installed.
#

TOPDIR=/opt/perforce/quick-start
HELIX_QUICK_START_DEFAULTS="${TOPDIR}/etc/defaults.cfg"
LOGFILE="/opt/perforce/quick-start/logs/install.log"

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  __hide
#   DESCRIPTION:  Hides text output from the console.
#------------------------------------------------------------------------------
__hide() {
    exec 4>&1 5>&2 >> "$LOGFILE" 2>&1
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  __show
#   DESCRIPTION:  Re-enables text output to the console.
#------------------------------------------------------------------------------
__show() {
    exec 1>&4 2>&5
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  __err_exit_not_fixable
#   DESCRIPTION:  Exit after printing any message passed in and path to log
#                 Intended to be called directly for irrecoverable errors
#------------------------------------------------------------------------------
__err_exit_not_fixable() {
    [ "$*" ] && echoerror "$*"
    echoerror "Further details may be in the log file at $LOGFILE"
    exit 1
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  __err_exit
#   DESCRIPTION:  Exit after printing an error message and how to resume
#------------------------------------------------------------------------------
__err_exit() {
    echoerror "$*"
    echoerror "Correct the error and then resume by running the following as root/sudo:
          /opt/perforce/quick-start/bin/configure-helix-quick-start.sh"
    __err_exit_not_fixable
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  __detect_color_support
#   DESCRIPTION:  Try to detect color support.
#------------------------------------------------------------------------------
__detect_color_support() {
    _COLORS=${BS_COLORS:-$(tput colors 2>/dev/null || echo 0)}
    if [ $? -eq 0 ] && [ "$_COLORS" -gt 2 ]; then
        RC="\033[1;31m"
        GC="\033[1;32m"
        BC="\033[1;34m"
        YC="\033[1;33m"
        EC="\033[0m"
    else
        RC=""
        GC=""
        BC=""
        YC=""
        EC=""
    fi
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  echoerr
#   DESCRIPTION:  Echo errors to stderr.
#------------------------------------------------------------------------------
echoerror() {
    [ "$_SUPPRESSED" -eq 1 ] && __show
    printf "${RC} * ERROR${EC}: %s\n" "$@" | tee -a "${LOGFILE}" 1>&2;
    [ "$_SUPPRESSED" -eq 1 ] && __hide
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  echoinfo
#   DESCRIPTION:  Echo information to stdout.
#------------------------------------------------------------------------------
echoinfo() {
    [ "$_SUPPRESSED" -eq 1 ] && __show
    printf "${GC} *  INFO${EC}: %s\n" "$@" | tee -a "${LOGFILE}";
    [ "$_SUPPRESSED" -eq 1 ] && __hide
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  echowarn
#   DESCRIPTION:  Echo warning informations to stdout.
#------------------------------------------------------------------------------
echowarn() {
    [ "$_SUPPRESSED" -eq 1 ] && __show
    printf "${YC} *  WARN${EC}: %s\n" "$@" | tee -a "${LOGFILE}";
    [ "$_SUPPRESSED" -eq 1 ] && __hide
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  echodebug
#   DESCRIPTION:  Echo debug information to stdout.
#------------------------------------------------------------------------------
echodebug() {
    if [ "$DEBUG" = "TRUE" ]; then
        # Show DEBUG info on screen, and record to the log
        [ "$_SUPPRESSED" -eq 1 ] && __show
        printf "${BC} * DEBUG${EC}: %s\n" "$@" | tee -a "${LOGFILE}";
        [ "$_SUPPRESSED" -eq 1 ] && __hide
    else
        # Only record DEBUG info to the log
        printf "${BC} * DEBUG${EC}: %s\n" "$@" >> "${LOGFILE}";
    fi
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  message
#   DESCRIPTION:  Output message text that the user needs to act upon.
#------------------------------------------------------------------------------
message() {
    printf "${GC}%s${EC}\n" "$@";
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  packages
#   DESCRIPTION:  Perform package configuration on Ubuntu to lock helix-p4d
#                 version in case of upgrade.
#------------------------------------------------------------------------------
packages() {
    P4D_VERSION="18.1"
    if [ -x "/usr/bin/apt-mark" ]
    then
        # Make sure that the current version of P4D both doesn't get
        # automatically removed on an upgrade, and that it is set as
        # the default even if a later version is installed.
        apt-mark manual "helix-p4d-base-$P4D_VERSION"
        update-alternatives --set helix-p4d "/opt/perforce/sbin/p4d.$P4D_VERSION"
    fi
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  get_system_info
#   DESCRIPTION:  Use lsb_release to determine OS and version.
#                 Then determine if it is supported.
#------------------------------------------------------------------------------
get_system_info() {
    # Let's test if the lsb_release binary is available, it should be
    if lsb_release >/dev/null 2>&1; then
        trusty=$(lsb_release -si)
        if echo "$trusty" | grep -q -i "RedHat" ; then
            # For the purposes of this script, we don't distinguish between RedHat and CentOS
            trusty='CentOS'
        fi
        DISTRO_VERSION=$(lsb_release -sr | cut -f1 -d".")
    else
        __err_exit "Could not find platform details using: [lsb_release]"
    fi

    # Determine if platoform is supported, exit if not
    if [ "$trusty" = 'Ubuntu' ]; then
        if [ [ "$DISTRO_VERSION" != "14" && [ "$DISTRO_VERSION" != "16" ] ; then
            __err_exit_not_fixable "This platform version is not supported."
        fi
    elif [ "$trusty" = 'CentOS' ]; then
        if [ "$DISTRO_VERSION" != "6" ] && [ "$DISTRO_VERSION" != "7" ] ; then
            __err_exit_not_fixable "This platform version is not supported."
        fi
    else
        __err_exit_not_fixable "This platform is not supported: [$trusty]"
    fi

    if uname >/dev/null 2>&1; then
    	DISTRO_PLATFORM=$(uname -i)
    	if [ "$DISTRO_PLATFORM" != "x86_64" ] ; then
    		__err_exit_not_fixable "64-bit Linux is required. This system appears to be [$DISTRO_PLATFORM]."
    	fi
	else
		__err_exit "Could not find platform details using: [uname]"
	fi

    # Attempt to get the system's IPv4 address
    # Imperfect, grabs the first one listed (typically eth0)
    IPV4_LIST=$(hostname -I)
    IPV4=$(echo $IPV4_LIST | cut -d ' ' -f1)
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  get_super_ticket
#   DESCRIPTION:  Use the default user and password values to set a ticket
#                 value and reusable p4 command prefix in the environment
#------------------------------------------------------------------------------
get_super_ticket () {

    # Don't rely on a $HOME in which to store a .p4tickets file
    # Get the ticket hash from login, then use it for future operations
    P4_SUPER_CMD="p4 -C utf8 -p ${P4PORT} -u ${P4_SUPER_USER}"
    TICKET="$(echo "${P4_SUPER_PASSWORD}" | ${P4_SUPER_CMD} login -p | egrep "^[0-9A-Z]{32}$" | head -n1 )"
    P4_SUPER_TICKET_CMD="${P4_SUPER_CMD} -P ${TICKET}"
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  configure_security
#   DESCRIPTION:  Ensure firewall ports are opened to the services we provide.
#                 On CentOS, also disable SELinux
#------------------------------------------------------------------------------
configure_security () {
    local helxmsg="Configuring firewall to allow Helix traffic (port 1666)"
    local httpmsg="Configuring firewall to allow HTTP traffic"
    if [ ${trusty} = 'Ubuntu' ]; then
        if ufw status 2>&1 | grep -q -e " active"; then
            echoinfo "$helxmsg"
            ufw allow 1666/tcp
            echoinfo "$httpmsg"
            ufw allow Apache
        fi
    elif [ ${trusty} = 'CentOS' ]; then
        if [[ ${DISTRO_VERSION} =~ 6 ]]; then
            if chkconfig --list iptables | grep -q -e "3:on"; then
                # lokkit automatically starts iptables if it is not running
                echoinfo "$helxmsg"
                lokkit -p 1666:tcp
                echoinfo "$httpmsg"
                lokkit -s http
            fi
        else
            # el7 uses systemd, assume later versions will as well
            if systemctl is-enabled firewalld; then
                # firewall-cmd fails if firewalld is not actually running
                systemctl start firewalld
                echoinfo "$helxmsg"
                firewall-cmd --permanent --add-port=1666/tcp
                echoinfo "$httpmsg"
                firewall-cmd --permanent --add-service=http
                systemctl reload firewalld
            fi
        fi
        echoinfo "Setting SELinux to 'permissive' mode"
        # for the current session
        setenforce 0 || true
        # persistently, so that it stays disabled after reboot
        [ -e /etc/selinux/config ] && sed -i "s/SELINUX=enforcing/SELINUX=permissive/" /etc/selinux/config
    else
        # should never get here.
        echoerror "No known firewall configuration steps for ${trusty}, you may need to take steps to allow Helix and HTTP traffic through your firewall."
    fi
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  configure_p4broker
#   DESCRIPTION:  Setup a p4broker to route external connections from ssl:1666
#                 to localhost:1667.
#------------------------------------------------------------------------------
configure_p4broker() {
    # shellcheck disable=SC2001
    local PORT=$(echo "${P4_BROKER_PORT}" | sed 's|.*:||')
    if p4dctl list -t p4broker 2>&1 | grep -q -w -e "${P4_BROKER_SERVICE_NAME}" ; then
        # This is the Broker instance that this script is responsible for
        echoinfo "A Helix Broker instance is already being managed by p4dctl with service name: [$P4_BROKER_SERVICE_NAME]"
    elif [ -d "$P4_BROKER_ROOT" ]; then
       # This is the Broker instance that this script is responsible for, but something is wrong
        echoerror "A Helix Broker instance already exists with root: [$P4_BROKER_ROOT],"
        echoerror "but it is not being managed by p4dctl with service name: [$P4_BROKER_SERVICE_NAME]"
        __err_exit "Configuration cannot continue in this state"
    elif echodebug "Looking for any process listening on the port required by Helix Broker: [${PORT}]"; nc localhost "${PORT}" < /dev/null; then
        # If something is listening here, it's not the Broker instance that this script is responsible for
        __err_exit "Another process appears to be listening on the port required by Helix Broker: [${PORT}]"
    else
        BROKER_CONFIG="$P4_BROKER_ROOT/broker.conf"
        P4DCTL_DIR="/etc/perforce/p4dctl.conf.d"
        BROKER_P4DCTL_CONFIG="$P4DCTL_DIR/$P4_BROKER_SERVICE_NAME.conf"

        echoinfo "Configuring Helix Broker"

        # Create a new broker configuration file, and configure it.
        mkdir -p "$P4_BROKER_ROOT"
        p4broker -C > "$BROKER_CONFIG"

        sed -i "s#\(target *= \).*#\1$P4PORT;#" "$BROKER_CONFIG"
        sed -i "s#\(listen *= \).*#\1$P4_BROKER_PORT;#" "$BROKER_CONFIG"
        sed -i "s#\(directory *= \).*#\1$P4_BROKER_ROOT;#" "$BROKER_CONFIG"

        # Now insert new broker rules.
        cat <<EOF >> "$BROKER_CONFIG"
command: ^password$|^passwd$
{
        args = ^${SWARM_ADMIN_USER}$;
        action = reject;
        message = "This action has been blocked by Helix Broker. Reason: changing the password for the '${SWARM_ADMIN_USER}' user will prevent the Swarm application from interacting with the Versioning Engine (P4D).";
}

command: ^user$
{
        args = ^${SWARM_ADMIN_USER}$;
        action = reject;
        message = "This action has been blocked by Helix Broker. Reason: the '${SWARM_ADMIN_USER}' user settings are managed by the Swarm application, and should not be modified separately.";
}

command: .*
{
        user = ^${SWARM_ADMIN_USER}$;
        action = reject;
        message = "This action has been blocked by Helix Broker. Reason: the '${SWARM_ADMIN_USER}' user is an application user that should not be used for interactive login.";
}

command: ^logout$
{
        args = ^${SWARM_ADMIN_USER}$;
        action = reject;
        message = "This action has been blocked by Helix Broker. Reason: logging out the '${SWARM_ADMIN_USER}' user will prevent the Swarm application from running.";
}

EOF

        # Generate the SSL certificates
        export P4SSLDIR="$P4_BROKER_ROOT/ssl"
        mkdir -p "$P4SSLDIR"
        chmod 700 "$P4SSLDIR"
        p4broker -Gc
        chown -R "$QUICK_START_USER:$QUICK_START_USER" "$P4_BROKER_ROOT"

        # Now configure the P4DCTL configuration for the broker.
        cp "$P4DCTL_DIR/p4broker.template" "$BROKER_P4DCTL_CONFIG"

        sed -i "s#%NAME%#$P4_BROKER_SERVICE_NAME#" "$BROKER_P4DCTL_CONFIG"
        sed -i "s#ssl#$P4SSLDIR#" "$BROKER_P4DCTL_CONFIG"

        chown "$QUICK_START_USER:$QUICK_START_USER" "$BROKER_P4DCTL_CONFIG"

        p4dctl start "$P4_BROKER_SERVICE_NAME"
    fi
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  configure_p4d
#   DESCRIPTION:  Configure p4d to listen to localhost:1667
#------------------------------------------------------------------------------
configure_p4d() {
    NO_SAMPLE_DATA_FLAG="$P4ROOT/no-sample-data.txt"
    # shellcheck disable=SC2001
    local PORT=$(echo "${P4PORT}" | sed 's|.*:||')
    if p4dctl list -t p4d 2>&1 | grep -q -w -e "${P4_SERVICE_NAME}" ; then
        # This is the P4D instance that this script is responsible for
        echoinfo "A Helix P4D instance is already being managed by p4dctl with service name: [$P4_SERVICE_NAME]"
    elif [ -d "$P4ROOT" ]; then
        # This is the P4D instance that this script is responsible for, but something is wrong
        echoerror "A Helix P4D instance already exists with root: [$P4ROOT],"
        echoerror "but it is not being managed by p4dctl with service name: [$P4_SERVICE_NAME]"
        __err_exit "Configuration cannot continue in this state"
    elif echodebug "Looking for any process listening on the port required by Helix P4D: [${PORT}]"; nc localhost "${PORT}" < /dev/null; then
        # If something is listening here, it's not the P4D instance that this script is responsible for
        __err_exit "Another process appears to be listening on the port required by Helix P4D: [${PORT}]"
    else
        HELIX_P4D_CONFIGURE_STRING="/opt/perforce/sbin/configure-helix-p4d.sh ${P4_SERVICE_NAME} -n -p ${P4PORT} -r ${P4ROOT} -u ${P4_SUPER_USER} -P ${P4_SUPER_PASSWORD} --case=${P4_CASE_MODE} --unicode"
        HELIX_P4D_P4DCTL_STATUS="p4dctl status ${P4_SERVICE_NAME}"

        echoinfo "Configuring Helix P4D"

        ${HELIX_P4D_CONFIGURE_STRING}
        if [ $? -ne 0 ]; then
            __err_exit "Helix P4D configuration failed"
        fi

        ${HELIX_P4D_P4DCTL_STATUS}
        if [ $? -ne 0 ]; then
            __err_exit "Helix P4D was configured, but is not running correctly, as reported by '${HELIX_P4D_P4DCTL_STATUS}'"
        else
            echodebug "Helix P4D has been configured successfully"

            # Leave a file to indicate this server was set up *without* any sample data,
            # so we can prevent a re-run of this script from accidentally overwriting important data with sample data
            [ ! "$INCLUDE_SAMPLE_DATA" = "TRUE" ] && echo "This P4D instance was created without any sample data by Helix Quick Start $(date)" > "$NO_SAMPLE_DATA_FLAG"
        fi
    fi
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  install_license
#   DESCRIPTION:  Install an expired license on the system. This must be the
#                 last step, since server commands will fail after this.
#------------------------------------------------------------------------------
install_license() {
    # If no license is present, set up null license
    if [ -f "$P4ROOT/license" ]; then
        echoinfo "A license file is already in place: [$P4ROOT/license]"
    else
        cp -pf "$TOPDIR/etc/license" "$P4ROOT/license"
        p4dctl restart "$P4_SERVICE_NAME"
    fi
}

configure_swarm() {
    SWARM_CONFIG="/opt/perforce/swarm/data/config.php"
    if [ -f "$SWARM_CONFIG" ]; then
        echoinfo "Helix Swarm is already configured"
    else
        # Perform all Swarm configuration
        echoinfo "Configuring Helix Swarm"

        echodebug "Starting sendmail, which will help Swarm send email notifications"
        # Start sendmail, let it create its keys
        service sendmail start

        if [ "$INCLUDE_SAMPLE_DATA" = "TRUE" ]
        then
            CREATE_USER=""
        else
            CREATE_USER="-c"
        fi

        # Prep a Swarm config file that will prevent the $SWARM_ADMIN_USER from logging in to the UI
        cat > $SWARM_CONFIG << __PREVENT_LOGIN__
<?php
return array(
    'security' => array(
        'prevent_login' => array(
            '${SWARM_ADMIN_USER}',
        ),
    ),
    'reviews' => array(
        'filters' => array (
            'result_sorting' => true,
            'date_field' => 'updated',
        ),
    ),
);
__PREVENT_LOGIN__

        # We will configure Swarm to use sendmail later, so the MX value is just a placeholder
        # Don't specify -H hostname, we want Swarm to pick the best hostname or FQDN
        HELIX_SWARM_MX="removeme"
        HELIX_SWARM_CONFIGURE_STRING="/opt/perforce/swarm/sbin/configure-swarm.sh -n -p ${P4PORT} -u ${SWARM_ADMIN_USER} -w ${SWARM_ADMIN_PASSWORD} -e ${HELIX_SWARM_MX} ${CREATE_USER} -g -U ${P4_SUPER_USER} -W ${P4_SUPER_PASSWORD} -B ${SWARM_BASE_URL}"

        ${HELIX_SWARM_CONFIGURE_STRING}
        if [ $? -ne 0 ]; then
            __err_exit "Helix Swarm configuration failed"
        else
            echodebug "Helix Swarm has been configured successfully"
        fi

        # Additional Swarm configuration steps

        # Configure Swarm's cron config to resolve a local address, in addtion to the value configure-swarm.sh picked for hostname
        echo -e "#\n# Configured by Quick Start for local resolution:\n127.0.0.1:80/${SWARM_BASE_URL}" >> /opt/perforce/etc/swarm-cron-hosts.conf

        # On empty P4D installations, create a //.swarm depot, used for Swarm attachments, etc.
        # Prefer a test for existence, in case one day Swarm install itself creates this depot
        if ${P4_SUPER_TICKET_CMD} depots | grep -q -e ".swarm"; then
            echodebug "Depot already exists: [//.swarm]"
        else
            echodebug "Creating depot: [//.swarm]"
            ${P4_SUPER_TICKET_CMD} depot -o ".swarm" | ${P4_SUPER_TICKET_CMD} depot -i
        fi

        # Set Swarm token and triggers
        echodebug "Configuring Swarm Triggers"

        # Sample data checkpoint already has triggers in it, don't add them a second time
        [ "$INCLUDE_SAMPLE_DATA" = "TRUE" ] || (${P4_SUPER_TICKET_CMD} triggers -o ; /opt/perforce/swarm/p4-bin/scripts/swarm-trigger.pl -o) | ${P4_SUPER_TICKET_CMD} triggers -i

        queue_dir="/opt/perforce/swarm/data/queue"
        token_dir="$queue_dir/tokens"
        mkdir -p $token_dir
        UUID=$(uuidgen)
        touch $token_dir/"$UUID"
        chmod 640 $token_dir/"$UUID"
        echo "SWARM_HOST=\"http://127.0.0.1/${SWARM_BASE_URL}\"" > /opt/perforce/etc/swarm-trigger.conf
        echo "SWARM_TOKEN=\"$UUID\"" >> /opt/perforce/etc/swarm-trigger.conf

        # Configure Swarm to use sendmail

        # Make config.php mods to use sendmail
        cp "$SWARM_CONFIG" "$SWARM_CONFIG.$(date +%Y%m%d_%H%M%S)"

        # Note that the default sendmail.cf file does not allow Sendmail to accept network connections from any host other than the local computer

        # Add the sender, so that php mail() does not attempt to use sendmail -f with the
        #  Swarm 'hostname', which may be an IP address, FQDN, or other alias,
        #  and not in /etc/mail/local-host-names
        sed -i "s|'mail'.*|&\n\t'sender' => 'notifications@localhost',|" $SWARM_CONFIG

        # Clear the mail rely host value, which causes Swarm to use sendmail instead
        sed -i "s|removeme||" $SWARM_CONFIG

        # Set property for P4V-Swarm integration, requires P4V >= 14.3
        echoinfo "Enabling Swarm integration for P4V"
        echodebug "Using URL: http://${IPV4}/${SWARM_BASE_URL}"
        ${P4_SUPER_TICKET_CMD} property -a -n P4.Swarm.URL -v http://${IPV4}/${SWARM_BASE_URL} || echowarn "Could not enable Swarm integration for P4V"

        # Platform specific Swarm configuration
        case ${trusty} in
           'Ubuntu')
                CRON_SERVICE="cron"
                chown -R www-data:www-data $queue_dir

                # Remove the virtual host and logs created by the Swarm configure script
                a2dissite perforce-swarm-site.conf
                rm -f /etc/apache2/sites-available/perforce-swarm-site.conf
                rm /var/log/apache2/swarm*

                # Disable default site if present
                for site in /etc/apache2/sites-enabled/*default*
                do
                    echoinfo "Disabling the default Apache site $site"
                    a2dissite "$( basename "$site" )"
                    #Apache restarted later to make this take effect
                done
                ;;
           'CentOS')
                CRON_SERVICE="crond"
                chown -R apache:apache $queue_dir

                # Remove the virtual host and logs created by the Swarm configure script
                rm -f /etc/httpd/conf.d/perforce-swarm-site.conf
                rm /var/log/httpd/swarm*
                ;;
           *)
                __err_exit_not_fixable "${trusty} not supported.";
                ;;
        esac

        # Apply Swarm configuration settings
        restart_apache

        # Cron must be running in order to start Swarm workers
        # Easiest way to ensure started regardless of starting state is to issue restart
        service ${CRON_SERVICE} restart || echowarn "The [${CRON_SERVICE}] service must be running in order for Swarm to work properly"
    fi

}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  restart_apache
#   DESCRIPTION:  Restart Apache and wait for it to be running.
#------------------------------------------------------------------------------
restart_apache() {

    if [ "${trusty}" = 'Ubuntu' ]; then
        APACHE_BINARY="apache2"
        APACHE_RESTART="apache2ctl restart"
        if [[ "${DISTRO_VERSION}" =~ 14 ]]; then
            IS_RUNNING_RE="pache2 is running" # Avoid case differences between versions
        elif [[ "${DISTRO_VERSION}" =~ 16 ]]; then
            IS_RUNNING_RE="running"
        fi
    elif [ "${trusty}" = 'CentOS' ]; then
        APACHE_BINARY="httpd"
        if [[ "${DISTRO_VERSION}" =~ 6 ]]; then
            APACHE_RESTART="service ${APACHE_BINARY} restart"
            IS_RUNNING_RE="httpd.*is running"
        elif [[ "${DISTRO_VERSION}" =~ 7 ]]; then
            APACHE_RESTART="systemctl restart ${APACHE_BINARY}"
            IS_RUNNING_RE="active \(running\)"
        else
            # should never get here.
            echowarn "No known Apache running status message on ${trusty} ${DISTRO_VERSION}, you may need to verify Apache is running."
        fi
    else
        # should never get here.
        __err_exit_not_fixable "Do not know how to treat Apache on ${trusty}"
    fi

    APACHE_STATUS="service ${APACHE_BINARY} status"

    ${APACHE_RESTART}
    retry=30
    until [[ $(${APACHE_STATUS}) =~ $IS_RUNNING_RE ]] ; do
        let retry-=1
        if [ $retry -gt 0 ]; then
	    echoinfo "Waiting for Apache to restart"
            sleep 1
        else
            echoerror "Apache2 failed to start. Configuration will continue, but a reboot is recommended."
            break
        fi
    done

}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  configure_apache
#   DESCRIPTION:  Set up a virtual host file to serve both the Quick Start and Swarm sites.
#------------------------------------------------------------------------------
configure_apache() {
    echoinfo "Configuring the Helix Quick Start and Swarm websites"
    if [ ${trusty} = 'Ubuntu' ]; then
        # Quiet warnings, "Could not reliably determine the server's fully qualified domain name"
        echo "ServerName localhost" >> /etc/apache2/apache2.conf

        a2enmod proxy proxy_http proxy_html
        if [ -f /etc/apache2/mods-available/xml2enc.load ]
        then
            # This is not available on Ubuntu 12.04
            a2enmod xml2enc
        fi

        APACHE_SITE_DIR="/etc/apache2/sites-available"

        # Apache on Ubuntu 12 doesn't use '.conf' extensions
        if [ "${DISTRO_VERSION}" = '12' ]
        then
            ln -sf "$APACHE_SITE_DIR/helix-quick-start.conf" "$APACHE_SITE_DIR/helix-quick-start"
        fi
        sed -i 's#APACHE_LOG_DIR_PLACEHOLDER#${APACHE_LOG_DIR}/access.log#' "$APACHE_SITE_DIR/helix-quick-start.conf"
        sed -i "s#WEBAPP_PORT#$WEBAPP_PORT#g" "$APACHE_SITE_DIR/helix-quick-start.conf"
        sed -i "s#SWARM_BASE_URL#$SWARM_BASE_URL#g" "$APACHE_SITE_DIR/helix-quick-start.conf"

        a2ensite helix-quick-start
    elif [ "${trusty}" = 'CentOS' ]; then
        # Quiet warnings, "Could not reliably determine the server's fully qualified domain name"
        echo "ServerName localhost" >> /etc/httpd/conf/httpd.conf

        # We don't have enabled/disabled options in CentOS,
        # so the Apache config is initially not deployed.
        APACHE_SITE_DIR="/etc/httpd/conf.d"

        sed -i 's#APACHE_LOG_DIR_PLACEHOLDER#logs/access_log#' "$TOPDIR/etc/helix-quick-start.conf"
        sed -i "s#WEBAPP_PORT#$WEBAPP_PORT#g" "$TOPDIR/etc/helix-quick-start.conf"
        sed -i "s#SWARM_BASE_URL#$SWARM_BASE_URL#g" "$TOPDIR/etc/helix-quick-start.conf"

        ln -sf $TOPDIR/etc/helix-quick-start.conf $APACHE_SITE_DIR/helix-quick-start.conf

        # We need this in order for the Apache proxy to work.
        #/usr/sbin/setsebool -P httpd_can_network_connect 1
    else
        # should never get here.
        __err_exit_not_fixable "Do not know how to treat Apache on ${trusty}"
    fi

    # Obtain the version of Apache we're using, and use it to enable/disable
    # access directives defined in the site template
    local apache_version="$(apachectl -v | grep '^Server version' | sed -e 's,.*Apache/\([0-9]\.[0-9]\).*,\1,')"
    [[ -n "$apache_version" ]] || __err_exit "Cannot determine Apache version"

    case "$apache_version" in
    '2.2')
        # Ubuntu 12 and CentOS 6 typically

        # Enable just the 2.2 directives, delete others
        sed -i 's/#APACHE2.2://' "$APACHE_SITE_DIR/helix-quick-start.conf"
        sed -i '/#APACHE.*:/d' "$APACHE_SITE_DIR/helix-quick-start.conf"
        ;;
    *)
        # Ubuntu 14 and CentOS 7 typically

        # The script wouldn't have gotten this far on platforms that use <2.2
        # If it's not 2.2, it's almost definitely 2.4, but assume that future
        # Apache versions will also use with the directives introduced in 2.4

        # Enable just the 2.4 directives, delete others
        sed -i 's/#APACHE2.4://' "$APACHE_SITE_DIR/helix-quick-start.conf"
        sed -i '/#APACHE.*:/d' "$APACHE_SITE_DIR/helix-quick-start.conf"
        ;;
    esac

    # Add a link from the DocumentRoot to the Swarm root
    (cd /var/www && ln -s /opt/perforce/swarm/public swarm)

    # Apache will be restarted after configuring Swarm
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  configure_sample_data
#   DESCRIPTION:  Install the sample depot data. We need to wipe out any
#                 existing configuration and rebuild from the checkpoint.
#------------------------------------------------------------------------------
configure_sample_data()
{
    SAMPLE_DATA_TAR="$TOPDIR/share/sample_data.tgz"
    SAMPLE_DATA_FLAG="$P4ROOT/sample-data.txt"

    if [ ! -f "$SAMPLE_DATA_TAR" ]; then
        __err_exit "Unable to find sample data archive: [$SAMPLE_DATA_TAR]"
    elif [ -f "$SAMPLE_DATA_FLAG" ]; then
        echoinfo "P4D has already been configured with sample data"
    elif [ -f "$NO_SAMPLE_DATA_FLAG" ]; then
        echoinfo "P4D was initially configured without sample data; refusing to overwrite potentially important user data"
    else
        echoinfo "Configuring Sample Data"

        p4dctl stop "$P4_SERVICE_NAME"

        # Ensure that we have a clean P4ROOT directory.
        if [ -d "$P4ROOT" ]
        then
            rm -fr "$P4ROOT/"*
        else
            mkdir -p "$P4ROOT"
        fi

        # Untar the checkpoint and archive files, and create a new database.
        tar -C "$P4ROOT" -xzf "$SAMPLE_DATA_TAR"
        checkpoint=$( find "$P4ROOT" -name "checkpoint.*.gz" | head -1 )
        /usr/sbin/p4d -r "$P4ROOT" -z -jr "$checkpoint"
        /usr/sbin/p4d -r "$P4ROOT" -xu
        chown -R "$QUICK_START_USER:$QUICK_START_USER" "$P4ROOT"

        # Start the new server
        p4dctl start "$P4_SERVICE_NAME"

        # Leave a flag so we can prevent a re-run of this script from accidentally
        # re-installing sample data without also doing other necessary configuration
        echo "This P4D instance was created from sample data provided by Helix Quick Start $(date)" > "$SAMPLE_DATA_FLAG"
    fi
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  customize_p4d
#   DESCRIPTION:  Custom post-configuration settings
#------------------------------------------------------------------------------
customize_p4d()
{
    # Ensure we can connect
    p4d_license_info="${P4_SUPER_TICKET_CMD} license -u"
    if ! $p4d_license_info || $p4d_license_info | grep -q -e 'License expired'; then
        # Shouldn't normally get here, as the de-license step also precludes re-configure
        echodebug "Helix P4D is either not accessible, or not licensed"
    else
        echodebug "Helix P4D is accessible"

        # Create an example Helix user for general evaluation.
        if $P4_SUPER_TICKET_CMD users 2>&1 | grep -q -e "$EXAMPLE_USER"; then
            echodebug "A Helix example user already exists: [${EXAMPLE_USER}]"
        else
            echodebug "Creating a Helix example user: [${EXAMPLE_USER}]"
            $P4_SUPER_TICKET_CMD user -o "$EXAMPLE_USER" | $P4_SUPER_TICKET_CMD user -fi
            echo -e "$EXAMPLE_USER_PASSWORD\n$EXAMPLE_USER_PASSWORD" | $P4_SUPER_TICKET_CMD passwd "$EXAMPLE_USER"
        fi

        # Enable DVCS connections
        if ${P4_SUPER_TICKET_CMD} configure set server.allowfetch=3 && ${P4_SUPER_TICKET_CMD} configure set server.allowpush=3 ; then
            echodebug "DVCS enabled"
        else
            echoerror "Helix P4D could not be configured to enable DVCS"
        fi

        # Enable structured logging
	if ${P4_SUPER_TICKET_CMD} configure set serverlog.file.1=commands.csv \
		&& ${P4_SUPER_TICKET_CMD} configure set serverlog.file.2=errors.csv \
		&& ${P4_SUPER_TICKET_CMD} configure set serverlog.file.3=auth.csv; then
            echodebug "Structured logging enabled"
        else
            echoerror "Helix P4D could not be configured for structured logging"
        fi
    fi
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  verify_configuration
#   DESCRIPTION:  Check basic functionality to verify configuration succeeded.
#------------------------------------------------------------------------------
verify_configuration()
{
    TEST_URL="127.0.0.1:80"
    if wget -nv -O - "${TEST_URL}" > /dev/null; then
        echoinfo "The Quick Start site is reachable locally"
    else
        echowarn "The Quick Start site is not reachable locally using: [${TEST_URL}]"
    fi

    TEST_URL="127.0.0.1:80/${SWARM_BASE_URL}/about"
    if wget -nv -O - --auth-no-challenge --user "${P4_SUPER_USER}" --password "${TICKET}" "${TEST_URL}" > /dev/null; then
        echoinfo "The Swarm site is reachable locally"
    else
        echowarn "The Swarm site is not reachable locally using: [${TEST_URL}]"
    fi
}

#---  FUNCTION  ---------------------------------------------------------------
#          NAME:  install_java_certs
#   DESCRIPTION:  Workaround to fix trustAnchors problem that can occur on
#                 versions prior to 16.04.
#------------------------------------------------------------------------------
install_java_certs()
{
    if [ "$DISTRO_VERSION" = "12" ] || [ "$DISTRO_VERSION" = "14" ] ; then
        /var/lib/dpkg/info/ca-certificates-java.postinst configure
    fi
}

#---  MAIN SCRIPT  ------------------------------------------------------------
#  DESCRIPTION:  Configures Helix Quick Start and its dependecies
#------------------------------------------------------------------------------
#

__detect_color_support
[ -d "$(dirname "$LOGFILE")" ] || mkdir -p "$(dirname "$LOGFILE")"
_SUPPRESSED=0

# No fancy messages, need to be able to run this before logging is set up
if [ "$(id -u)" -ne 0 ]
then
    echo "This script needs root privileges, please either login as root or use sudo."
    exit 1
fi

if [ -e ${HELIX_QUICK_START_DEFAULTS} ]
then
    echoinfo "Sourcing Configuration Defaults from ${HELIX_QUICK_START_DEFAULTS}"
    . ${HELIX_QUICK_START_DEFAULTS}
else
    __err_exit "The Helix Quick Start configuration file cannot be found at ${HELIX_QUICK_START_DEFAULTS}"
fi

if [ -f "${TOPDIR}/etc/configured" ]
then
    __err_exit_not_fixable "The Helix Quick Start system has already been configured."
fi

if [ "$1" = "--sampledata" ]
then
    INCLUDE_SAMPLE_DATA="TRUE"
    if [ "$(grep INCLUDE_SAMPLE_DATA "$HELIX_QUICK_START_DEFAULTS")" ]
    then
        sed -i 's/INCLUDE_SAMPLE_DATA=.*/INCLUDE_SAMPLE_DATA=TRUE/g' "$HELIX_QUICK_START_DEFAULTS"
    else
        echo "INCLUDE_SAMPLE_DATA=TRUE" >> "$HELIX_QUICK_START_DEFAULTS"
    fi
fi

__hide
_SUPPRESSED=1

get_system_info

configure_security

packages

configure_apache

configure_p4d

if [ "$INCLUDE_SAMPLE_DATA" = "TRUE" ]
then
    # Customizations are built into the sample data
    configure_sample_data
    get_super_ticket
else
    get_super_ticket
    customize_p4d
fi

configure_swarm

configure_p4broker

service helix-quick-start restart || __err_exit "Helix Quick Start failed to start."

verify_configuration

install_license

install_java_certs

echo "1" > "${TOPDIR}/etc/configured"

_SUPPRESSED=0
__show

message ""
message "::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::"
message "::"
message "::  Helix Quick Start has been configured, and includes:"
message "::"
message "::  - Helix Versioning Engine (P4D)"
message "::    - Accompanied by Helix P4Broker and P4Dctl"
message "::  - Helix Swarm"
message "::  - Helix Quick Start web application"
[ "$INCLUDE_SAMPLE_DATA" = "TRUE" ] && message "::  - Sample Data"
message "::"
message "::  In order to begin your trial, please point your web browser at:"
message "::"
message "::    http://<address-of-this-host>/helix"
message "::"
message "::  The address may be a Fully Qualified Domain Name, or an IP."
message "::  This host appears to use the following IP(s):"
for ip in $IPV4_LIST; do
    message "::    $ip"
done;
message "::"
message "::  If you need help, contact support@perforce.com"
message "::"
message "::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::"
message ""

exit 0
