#!/bin/bash
# Stop shellcheck alerting log redirects
# shellcheck disable=SC2094
#
TRUE=1
FALSE=0
__BASEDIR=$(dirname "$(dirname "$0")")
__RUNDATE=$(date +"%Y%m%d")
LOGFILE="${__BASEDIR}/logs/daily-tasks.${__RUNDATE}.log"
[ -d "$(dirname "$LOGFILE")" ] || mkdir -p "$(dirname "$LOGFILE")"

#
#---  FUNCTION  -------------------------------------------------------------------------------------------------------
#          NAME:  __detect_color_support
#   DESCRIPTION:  Try to detect color support
#----------------------------------------------------------------------------------------------------------------------
__detect_color_support() {
    if [ "$(tput colors 2>/dev/null || echo 0)" -gt 2 ]; then
        RC="\033[1;31m"
        GC="\033[1;32m"
        BC="\033[1;34m"
        YC="\033[1;33m"
        BOLD="\033[1m"
        EC="\033[0m"
    else
        RC=""
        GC=""
        BC=""
        YC=""
        BOLD=""
        EC=""
    fi
}
#---  FUNCTION  -------------------------------------------------------------------------------------------------------
#          NAME:  msg
#   DESCRIPTION:  Echo message to stdout
#----------------------------------------------------------------------------------------------------------------------
echomsg() {
    printf "${BOLD}%s${EC}\n" "$@" | tee -a "$LOGFILE";
    return 0
}
#---  FUNCTION  -------------------------------------------------------------------------------------------------------
#          NAME:  echoerr
#   DESCRIPTION:  Echo errors to stderr
#----------------------------------------------------------------------------------------------------------------------
echoerror() {
    printf "${RC} * ERROR${EC}: %s\n" "$@" | tee -a "$LOGFILE" 1>&2;
    return 0
}
#---  FUNCTION  -------------------------------------------------------------------------------------------------------
#          NAME:  echoinfo
#   DESCRIPTION:  Echo information to stdout
#----------------------------------------------------------------------------------------------------------------------
echoinfo() {
    printf "${GC} *  INFO${EC}: %s\n" "$@" | tee -a "$LOGFILE";
    return 0
}
#---  FUNCTION  -------------------------------------------------------------------------------------------------------
#          NAME:  echowarn
#   DESCRIPTION:  Echo warning to stdout
#----------------------------------------------------------------------------------------------------------------------
echowarn() {
    printf "${YC} *  WARN${EC}: %s\n" "$@" | tee -a "$LOGFILE";
    return 0
}

#---  FUNCTION  -------------------------------------------------------------------------------------------------------
#          NAME:  echodebug
#   DESCRIPTION:  Echo errors to stderr
#----------------------------------------------------------------------------------------------------------------------
echodebug() {
    [ "${DEBUG}" -eq ${TRUE} ] && printf "${BC} * DEBUG${EC}: %s\n" "$@" | tee -a "$LOGFILE" 1>&2;
    return 0
}

#---  FUNCTION  -------------------------------------------------------------------------------------------------------
#          NAME:  echodebug
#   DESCRIPTION:  Echo errors to stderr
#----------------------------------------------------------------------------------------------------------------------
teelog() {
    echo "$*" | sed 
    return 0
}

#---  FUNCTION  -------------------------------------------------------------------------------------------------------
#          NAME:  validate_environment
#   DESCRIPTION:  Check that p4dctl is consistent with the service configuration
#----------------------------------------------------------------------------------------------------------------------
__validate_environment() {
    echoinfo "Checking that the service is configured as expected."
    echodebug "Getting service information from p4dctl"
    if echodebug "Checking that p4dctl knows about ${P4_SERVICE_NAME}" && \
      ! p4dctl list -t p4d 2>&1 | tee -a "$LOGFILE" | grep "^[^\s]*\s*[^\s]*\s*${P4_SERVICE_NAME}" >/dev/null; then
      echoerror "Failed to identify ${P4_SERVICE_NAME} as a p4d instance."
      return 1
    fi
    SERVICE_ENV="$(p4dctl env "${P4_SERVICE_NAME}" P4ROOT 2>>"$LOGFILE" | tee -a "$LOGFILE" )"
    if [ -z "${SERVICE_ENV}" ]; then
      echoerror "Failed to identify P4ROOT for ${P4_SERVICE_NAME}"
      return 1
    elif echodebug "Checking for P4ROOT, ${SERVICE_ENV#P4ROOT=}" && \
      [ ! -d "${SERVICE_ENV#P4ROOT=}" ]; then
      echoerror "${SERVICE_ENV#P4ROOT=} does not exist."
      return 1;
    elif echodebug "Looking for database files, ${SERVICE_ENV#P4ROOT=}/db.*" && \
      [ ! "$(find "${SERVICE_ENV#P4ROOT=}" -depth -follow -name "db.*" | wc -l )" -gt 0 ]; then
      echoerror "There are no db files in ${SERVICE_ENV#P4ROOT=}."
      return 1
    elif echodebug "Looking for journalPrefix, ${SERVICE_ENV#P4ROOT=}" && \
      p4d -r "${SERVICE_ENV#P4ROOT=}" -cshow 2>>"$LOGFILE" | tee -a "$LOGFILE" | grep -i journalPrefix > /dev/null; then
      echoerror "journalPrefix has been set for ${P4_SERVICE_NAME}."
      return 1
    fi
    echodebug "$P4_SERVICE_NAME is suitable for the daily tasks"
    return 0
}
#---  FUNCTION  -------------------------------------------------------------------------------------------------------
#          NAME:  __checkpoint_service
#   DESCRIPTION:  Perform a p4dctl level checkpoint of a service, unless this is a dry-run.
#----------------------------------------------------------------------------------------------------------------------
__checkpoint_service(){
	trap 'rm -f /tmp/timestamp 2>>"$LOGFILE"' EXIT
    if [ -z "$1" ]; then
      echoerror "No service was passed in to the checkpoint function."
    elif [ "$_DRY_RUN" == "${TRUE}" ]; then
      echowarn "Dry-run mode, not checkpointing"
      return 0           
    fi
    # p4dctl reports failure when it has to use p4d -jc (job088506)
    touch /tmp/timestamp 2>>"$LOGFILE" && sleep 1
    if ! p4dctl checkpoint "$1" >>"$LOGFILE" 2>&1; then
       echowarn "p4dctl reported a checkpoint failure, looking for a new checkpoint file"
       # Try to work out whether a checkpoint was taken
       SERVICE_ENV="$(p4dctl env "${P4_SERVICE_NAME}" P4ROOT)"
       if [ ! "$(find "${SERVICE_ENV#P4ROOT=}" -depth -follow -newer /tmp/timestamp \( -name 'checkpoint\.[0-9]*' -o -name 'journal\.[0-9]*' \) | wc -l )" -eq 3 ]; then
         echoerror "p4dctl reported a failure and there were less than 3 new checkpoint/journal files"
         return 1
       fi
    fi
    return 0
}
#---  FUNCTION  -------------------------------------------------------------------------------------------------------
#          NAME:  __remove_old_checkpoints
#   DESCRIPTION:  Remove checkpoint and journal files older than _CHECKPOINT_MAX_AGE, unless this is a dry-run.
#----------------------------------------------------------------------------------------------------------------------
__delete_old_checkpoints(){
    if [ -z "$1" ]; then
      echoerror "No service was passed in to the remove old checkpoints function."
      return 1
    fi
    # Deleting stuff, so be sure about p4root
    SERVICE_ENV="$(p4dctl env "${1}" P4ROOT 2>>"$LOGFILE" | tee -a "$LOGFILE" )"
    if [ -z "${SERVICE_ENV}" ]; then
      echoerror "Failed to identify P4ROOT for ${P4_SERVICE_NAME}"
      return 1
    elif echodebug "Checking for P4ROOT, ${SERVICE_ENV#P4ROOT=}" && \
      [ ! -d "${SERVICE_ENV#P4ROOT=}" ]; then
      echoerror "${SERVICE_ENV#P4ROOT=} does not exist."
      return 1;
    fi
    OLD_FILES="$(find "${SERVICE_ENV#P4ROOT=}" -depth -type f \
        \( -name 'checkpoint\.[0-9]*' -o -name 'journal\.[0-9]*' \) \
        -mtime +"${_MAX_CHECKPOINT_AGE:-0}" 2>>"$LOGFILE" | tee -a "$LOGFILE")"
    if [ "$_DRY_RUN" == "${TRUE}" ]; then
      echowarn "Dry-run mode, not removing ${OLD_FILES:-anything(no old files anyway)}"
    elif [ -z "${OLD_FILES}" ]; then
      echowarn "There are no checkpoint/journal files older than ${_MAX_CHECKPOINT_AGE} days."
    elif ! echo -e ${OLD_FILES} | xargs rm >"$LOGFILE" 2>&1; then
      echoerror "Failed to remove old checkpoint/journal files, ${OLD_FILES}"
      return 1
    fi
    return 0
}
#---  FUNCTION  -------------------------------------------------------------------------------------------------------
#          NAME:  __usage
#   DESCRIPTION:  Exit after printing a usage statement
#----------------------------------------------------------------------------------------------------------------------
usage() {
     
    echomsg "--checkpoint             Checkpoint the perforce server, default false"
    echomsg "--max-checkpoint-age nn  How many days to retain checkpoint/journal files, 0 - infinite"
    echomsg "-n|--dry-run             Only validate and report on what would be done"
    echomsg "-v|--verbose             Turn debug on, also works with DEBUG=\"TRUE\" in the environment"
    echomsg "-V|--version             Print the script version and exit"
    echomsg "-?|-h|--help             Print the usage and exit"
    echomsg "<P4_SERVICE_NAME>        The name of the server being managed."
    exit 1
}
#
#---  MAIN SCRIPT  ----------------------------------------------------------------------------------------------------
#  DESCRIPTION:  Runs any maintenance tasks; checkpointing or filesystem tidy up
#----------------------------------------------------------------------------------------------------------------------
#
# Default options
#
#----------------------------------------------------------------------------------------------------------------------
#
__detect_color_support

P4_SERVICE_NAME=
_TAKE_CHECKPOINT=0
_MAX_CHECKPOINT_AGE=0
_DRY_RUN=0
# Translate DEBUG="TRUE" in the environment to the numerical value used here
[ "$DEBUG" = "TRUE" ] && DEBUG=${TRUE}
[ "$DEBUG" = "${TRUE}" ] || DEBUG=${FALSE}

#----------------------------------------------------------------------------------------------------------------------
#
# Parse command line options
#
#    --checkpoint             Checkpoint the perforce server, default false
#    --max-checkpoint-age nn  How many days to retain checkpoint/journal files, 0 - infinite
#    -v|--verbose             Turn debug on, also works with DEBUG="TRUE" in the environment
#    -V|--version             Print the script version and exit
#    -?|-h|--help             Print the usage and exit
#    <P4_SERVICE_NAME>        The name of the server being managed.
#
#----------------------------------------------------------------------------------------------------------------------
#
while true; do
  case "$1" in
    --checkpoint ) 
        _TAKE_CHECKPOINT="${TRUE}"; echodebug "Checkpoint enabled";
        shift;;
    --max-checkpoint-age ) 
        _MAX_CHECKPOINT_AGE="$2"; echodebug "Checkpoints will be retained for ${_MAX_CHECKPOINT_AGE} days, where 0 means for ever.";
        shift 2;;
     -n | --dry-run )
        _DRY_RUN="${TRUE}"; echowarn "This is a dry-run, no tasks will be performed";
        shift;;
     -v | --verbose )
        DEBUG="${TRUE}"; echodebug "Debug enabled"; echodebug "Invoked with $*";
        shift;;
     -h | -\? | --help ) usage;
        shift;;
    -* ) echoerror "unknown option $1";usage;
        shift;;
    * ) P4_SERVICE_NAME="$1"; echodebug "Finished with flags";
        break ;;
  esac
done
# Check that a service name has been provided
if [ -z "${P4_SERVICE_NAME}" ]; then
  echoerror "No service name provided"
  exit 1
fi

# Validate environment
if ! __validate_environment; then
  echoerror "Failed to confirm that daily tasks can be carried out."
  exit 1
fi

# Checkpoint the server
if [ "${_TAKE_CHECKPOINT}" -eq ${TRUE} ] ; then
    echoinfo "Checkpointing ${P4_SERVICE_NAME}"
    if __checkpoint_service "$P4_SERVICE_NAME"; then
      echoinfo "Checkpoint of ${P4_SERVICE_NAME} complete"
    else
      echoerror "Failed to checkpoint ${P4_SERVICE_NAME}"
      exit 1
    fi
else
    echowarn "Checkpointing has not been requested, no --checkpoint option supplied."
fi

# Tidy up old checkpoints and journals
if [ "${_MAX_CHECKPOINT_AGE}" -gt 0 ] ; then
    echoinfo "Deleting checkpoint and journal files older than ${_MAX_CHECKPOINT_AGE} days."
    if __delete_old_checkpoints "$P4_SERVICE_NAME"; then
        echoinfo "Old checkpoint deletion for ${P4_SERVICE_NAME} is complete."
    else
        echoerror "Failed to delete old checkpoints for ${P4_SERVICE_NAME}."
        exit 1
    fi
else
    echowarn "Checkpoints are allowed to age indefinitely."
fi

echoinfo "Daily tasks completed sucessfully"
